from pathlib import Path

import joblib
from huggingface_hub import hf_hub_download

REPO_MAP = {
    "hubert-base": "yfyeung/kmeans-librispeech-hubert-base",
    "wavlm-base": "yfyeung/kmeans-librispeech-wavlm-base",
}


def download_kmeans_model(model_type, layer, vocab):
    """Download k-means model from HuggingFace if not cached."""
    if model_type not in REPO_MAP:
        raise ValueError(f"Unsupported model_type: {model_type}")

    repo_id = REPO_MAP[model_type]
    filename = f"{model_type}-layer-{layer}.kmeans.{vocab}.model"

    cache_dir = Path.home() / ".cache" / "dswed" / "kmeans"
    cache_dir.mkdir(parents=True, exist_ok=True)

    model_path = hf_hub_download(
        repo_id=repo_id, filename=filename, cache_dir=str(cache_dir)
    )

    return joblib.load(model_path)



import argparse
import logging
from argparse import Namespace

import torch
from Levenshtein import distance as levenshtein_distance
from silero_vad import get_speech_timestamps, load_silero_vad

from .kmeans_loader import download_kmeans_model
from .models import get_model


def int_array_to_chinese_unicode(arr):
    """
    Map each integer value in the array to a distinct Unicode Chinese character.
    Unicode region for Chinese characters: 4E00 - 9FFF (20992 characters)

    Args:
        arr (list): Array of integers.
    Returns:
        str: Unicode Chinese sentence.
    """
    # Base Unicode point for Chinese characters.
    base_unicode_point = 0x4E00

    # Convert each integer in the array to a Unicode Chinese character.
    unicode_sentence = "".join(chr(base_unicode_point + val) for val in arr)

    return unicode_sentence


class DSWED:
    def __init__(
        self,
        model_type="hubert-base",
        vocab=50,
        layer=-1,
    ):
        """
        Args:
            model_type (str): Model type. Select from "hubert-base (int): Number of vocabularies for k-means clustering. Select from 50, 100, 200.
            vocab (int): Number of vocabularies for k-means clustering. Select from 50, 100, 200.
            layer (int): Layer number to extract features. If None, the last layer is used.
        """
        formatter = "%(asctime)s %(levelname)s [%(filename)s:%(lineno)d] %(message)s"
        logging.basicConfig(format=formatter, level=logging.INFO)

        model_name, model_version = model_type.split("-")
        self.args = Namespace(
            model_name=model_name,
            model_version=model_version,
            layer_idx=layer,
            n_clusters=vocab,
        )
        logging.info(vars(self.args))

        self.device = "cuda" if torch.cuda.is_available() else "cpu"

        self.model = get_model(self.args)
        self.model.eval()

        logging.info(
            f"Loading k-means model: {model_type}-layer-{layer}.kmeans.{vocab}"
        )
        self.km_model = download_kmeans_model(model_type, layer, vocab)
        self.vad_model = load_silero_vad()

    def get_parser(self):
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )

        parser.add_argument(
            "--model-name",
            type=str,
            default="hubert",
            choices=["hubert", "wavlm"],
        )

        parser.add_argument(
            "--model-version",
            type=str,
            required=True,
        )

        parser.add_argument(
            "--layer-idx",
            type=int,
            default=8,
            help="The index starts from 1, so if you want the 12-th layer feature, just set it to 12",
        )

        parser.add_argument(
            "--n-clusters",
            type=int,
            default=50,
        )

        return parser.parse_args()

    def trim_silence(self, wav):
        timestamps = get_speech_timestamps(wav, self.vad_model)
        if len(timestamps) == 0:
            return wav
        start = timestamps[0]["start"]
        end = timestamps[-1]["end"]
        return wav[start:end]

    @torch.no_grad()
    def decode_label(self, audio):
        """
        Args:
            audio (torch.Tensor): Audio waveform tensor (1, T).
        Returns:
            list: List of integers.
        """
        audio = audio.to(self.device)
        batch = {"audio": audio, "audio_lens": torch.LongTensor([audio.shape[1]])}
        layer_results, embedding_lens = self.model.extract_features(
            batch, self.args.layer_idx
        )
        embedding = layer_results[0, : embedding_lens[0]]
        km_label = self.km_model.predict(embedding).tolist()
        return km_label

    def score(self, gt_wav, gen_wav, weights=(1, 1, 1)):
        """
        Args:
            gt_wav (np.ndarray): Ground truth waveform (T,).
            gen_wav (np.ndarray): Generated waveform (T,).
        Returns:
            float: Distance score.
        """
        gt_wav = torch.from_numpy(gt_wav).float()
        gen_wav = torch.from_numpy(gen_wav).float()

        gt_wav = self.trim_silence(gt_wav)
        gen_wav = self.trim_silence(gen_wav)

        gt_wav = gt_wav.unsqueeze(0).to(self.device)
        gen_wav = gen_wav.unsqueeze(0).to(self.device)

        gt_label = self.decode_label(gt_wav)
        gen_label = self.decode_label(gen_wav)
        gt_text = int_array_to_chinese_unicode(gt_label)
        gen_text = int_array_to_chinese_unicode(gen_label)
        dist_score = levenshtein_distance(gen_text, gt_text, weights=weights)
        return dist_score



import argparse

import torch


def remove_long_utterances(c):
    if c.duration > 30.0:
        return False
    return True


def make_pad_mask(lengths: torch.Tensor, max_len: int = 0) -> torch.Tensor:
    """
    Args:
      lengths:
        A 1-D tensor containing sentence lengths.
      max_len:
        The length of masks.
    Returns:
      Return a 2-D bool tensor, where masked positions
      are filled with `True` and non-masked positions are
      filled with `False`.

    >>> lengths = torch.tensor([1, 3, 2, 5])
    >>> make_pad_mask(lengths)
    tensor([[False,  True,  True,  True,  True],
            [False, False, False,  True,  True],
            [False, False,  True,  True,  True],
            [False, False, False, False, False]])
    """
    assert lengths.ndim == 1, lengths.ndim
    max_len = max(max_len, lengths.max())
    n = lengths.size(0)
    seq_range = torch.arange(0, max_len, device=lengths.device)
    expaned_lengths = seq_range.unsqueeze(0).expand(n, max_len)

    return expaned_lengths >= lengths.unsqueeze(-1)


def str2bool(v):
    """Used in argparse.ArgumentParser.add_argument to indicate
    that a type is a bool type and user can enter

        - yes, true, t, y, 1, to represent True
        - no, false, f, n, 0, to represent False

    See https://stackoverflow.com/questions/15008758/parsing-boolean-values-with-argparse  # noqa
    """
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")



import torch
from transformers import AutoModel, Wav2Vec2FeatureExtractor


def get_model(args):
    model_name = args.model_name
    if model_name == "wavlm":
        model = WavLMModel(model_version=args.model_version)
    elif model_name == "hubert":
        model = HuBERTModel(model_version=args.model_version)
    else:
        raise ValueError(f"{model_name} is not supported yet")

    return model


class HuBERTModel(torch.nn.Module):
    def __init__(self, model_version: str = "base"):
        super().__init__()
        if model_version == "large":
            self.processor = Wav2Vec2FeatureExtractor.from_pretrained(
                f"facebook/hubert-{model_version}-ll60k"
            )
            self.model = AutoModel.from_pretrained(
                f"facebook/hubert-{model_version}-ll60k"
            )
        elif model_version == "base":
            self.processor = Wav2Vec2FeatureExtractor.from_pretrained(
                f"facebook/hubert-{model_version}-ls960"
            )
            self.model = AutoModel.from_pretrained(
                f"facebook/hubert-{model_version}-ls960"
            )
            self.processor.do_normalize = False
        else:
            raise ValueError(f"Unseen model version: {model_version}")

    def extract_features(self, batch, layer_idx, return_numpy=True):
        audio_pt = batch["audio"]
        if isinstance(audio_pt, list):
            audios = [audio.numpy() for audio in audio_pt]
        else:
            audio_lens_pt = batch["audio_lens"]
            audios = []
            for i in range(audio_pt.shape[0]):
                audios.append(audio_pt[i, : audio_lens_pt[i]].tolist())

        device = next(self.model.parameters()).device
        inputs = self.processor(
            audios,
            sampling_rate=16000,
            padding=True,
            return_attention_mask=True,
            return_tensors="pt",
        ).to(device)

        outputs = self.model(output_hidden_states=True, **inputs)
        layer_results = outputs.hidden_states[layer_idx]
        padding_mask = self.model._get_feature_vector_attention_mask(
            layer_results.shape[1], inputs["attention_mask"]
        )
        embedding_lens = padding_mask.sum(dim=1)

        if return_numpy:
            layer_results = layer_results.cpu().numpy()
        return layer_results, embedding_lens


class WavLMModel(torch.nn.Module):
    def __init__(self, model_version: str = "base"):
        super().__init__()
        from transformers import WavLMModel as HFWavLM

        model_name = f"microsoft/wavlm-{model_version}"
        self.model = HFWavLM.from_pretrained(model_name)
        self.processor = Wav2Vec2FeatureExtractor.from_pretrained(model_name)

    def extract_features(self, batch, layer_idx):
        audio_pt = batch["audio"]
        if isinstance(audio_pt, list):
            audios = [audio.numpy() for audio in audio_pt]
        else:
            audio_lens_pt = batch["audio_lens"]
            audios = []
            for i in range(audio_pt.shape[0]):
                audios.append(audio_pt[i, : audio_lens_pt[i]].tolist())

        device = next(self.model.parameters()).device
        inputs = self.processor(
            audios,
            sampling_rate=16000,
            padding=True,
            return_attention_mask=True,
            return_tensors="pt",
        ).to(device)

        outputs = self.model(output_hidden_states=True, **inputs)
        layer_results = outputs.hidden_states[layer_idx]
        padding_mask = self.model._get_feature_vector_attention_mask(
            layer_results.shape[1], inputs["attention_mask"]
        )
        embedding_lens = padding_mask.sum(dim=1)

        return layer_results.cpu().numpy(), embedding_lens



from setuptools import find_packages, setup

setup(
    name="dswed",
    version="1.0.0",
    packages=find_packages(),
    install_requires=[
        "numpy>=2.0.0",
        "transformers>=4.36.2",
        "torch>=2.0.0",
        "torchaudio>=0.10.1",
        "joblib>=1.0.1",
        "Levenshtein>=0.27.0",
        "huggingface_hub>=0.20.0",
        "scikit-learn>=1.0.0",
        "silero-vad>=5.0.0",
    ],
    author="Yifan Yang",
    author_email="yifanyeung@sjtu.edu.cn",
    description="A package for computing DS-WED.",
    url="https://github.com/yfyeung/DS-WED",
    keywords="speech metrics",
)
